/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Algorithm for flow assignment problems which routes all traffic of each 
 * demand, through the shortest path, and then creates a backup path, measured 
 * in number of hops or in km, being this a user-defined parameter.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class FA_shortestPath_11pathProtection implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		int N = netPlan.getNumberOfNodes();
		int E = netPlan.getNumberOfLinks();
		int D = netPlan.getNumberOfDemands();
		if (N == 0 || E == 0 || D == 0) throw new Net2PlanException("This algorithm requires a topology and a demand set");

		/* Initialize some variables */
		String shortestPathType = algorithmParameters.get("shortestPathType");
		if (!shortestPathType.equalsIgnoreCase("km") && !shortestPathType.equalsIgnoreCase("hops"))
			throw new Net2PlanException("Wrong shortestPathType parameter");

		Set<Long> linkIds = netPlan.getLinkIds();
		Map<Long, Double> linkCostMap = shortestPathType.equalsIgnoreCase("hops") ? DoubleUtils.ones(linkIds) : netPlan.getLinkLengthInKmMap();
		
		/* Remove all routing information */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);

		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
		Set<Long> demandIds = netPlan.getDemandIds();
		for(long demandId : demandIds)
		{
			long a_d = netPlan.getDemandIngressNode(demandId);
			long b_d = netPlan.getDemandEgressNode(demandId);
			double h_d = netPlan.getDemandOfferedTraffic(demandId);
			
			/* First compute all the primary and backup paths */
			/* If for two nodes, we could not find primary & backup paths => return */
			List<Long> primaryPath = GraphUtils.getShortestPath(linkMap, a_d, b_d, linkCostMap);
			if (primaryPath.isEmpty())
			{
				throw new Net2PlanException("The network is not connected");
			}
			
			Map<Long, Double> onesInNonUsedLinksMap = DoubleUtils.ones(linkIds);
			for (long linkId : primaryPath)
			{
				onesInNonUsedLinksMap.put(linkId, 0.0);
			}
			
			List<Long> backupPath = GraphUtils.getCapacitatedShortestPath(linkMap, a_d, b_d, linkCostMap, onesInNonUsedLinksMap, 1);
			if (backupPath.isEmpty())
			{
				throw new Net2PlanException("Could not find a link disjoint backup path");
			}
			
			/* Update the netPlan object with the routes and the protection segments */
			long routeId = netPlan.addRoute(demandId, h_d, primaryPath, null);
			long backupSegment = netPlan.addProtectionSegment(backupPath, h_d, null);
			netPlan.addProtectionSegmentToRouteBackupSegmentList(backupSegment, routeId);
		}
		
		return "Ok!";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> parameters = new LinkedList<Triple<String, String, String>>();
		parameters.add(Triple.of("shortestPathType", "#select# hops km", "Each demand is routed according to the shortest path according to this type. Can be 'km' or 'hops'"));
		
		return parameters;
	}

	@Override
	public String getDescription()
	{
		return "Algorithm for flow assignment problems which routes all traffic of each demand, through the shortest path, and"
				+ "then creates a backup path , measured in number of hops or in km, being this a user-defined parameter.";
	}
}
